/* yams.c -- Yet Another Machine Simulator
   Copyright (C) 2002 Juha Aatrokoski, Timo Lilja, Leena Salmela,
     Teemu Takanen, Aleksi Virtanen

   Copyright (C) 1992, 1995, 1996, 1997-1999, 2000-2002
                  Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.

   $Id: yams.c,v 1.20 2002/11/26 10:40:24 ttakanen Exp $
*/

/* AIX requires this to be the first thing in the file.  */
#if defined (_AIX) && !defined (__GNUC__)
 #pragma alloca
#endif

#include "includes.h"
#include "simulator.h"
#include "io.h"
#include "hwconsole.h"
#include "cfg.h"
#include "memory.h"

struct option longopts[] =
   {{ "config", required_argument, NULL, 'c'},
    { "help", no_argument, NULL, 'h' },
    { "script", required_argument, NULL, 's'},
    { "version", no_argument, NULL, 'v' },
    { NULL, 0, NULL, 0 }};

extern char version[];

int main(int argc, char *argv[])
{
  int optc;
  int h = 0, v = 0, c = 0, s = 0, binary = 0, lose = 0;
  int exit_code = 0, scripts = 0;

  char config_file[MAX_FILENAME_LENGTH];
  char script_file[MAX_FILENAME_LENGTH][MAX_SCRIPTS];
  char *binary_file;
  char kernel_params[KERNEL_PARAMAREA_LENGTH];

    while ((optc = getopt_long(argc, argv, "+c:hs:v", longopts, (int *) 0))
           != EOF)

      switch (optc) {
      case 'v':
          v = 1;
          break;
      case 'h':
          h = 1;
          break;
      case 'c':
          c = 1;
          strncpy(config_file, optarg, MAX_FILENAME_LENGTH);
          break;
      case 's':
          s = 1;
          if (scripts < MAX_SCRIPTS)
              strncpy(script_file[scripts++], optarg, MAX_FILENAME_LENGTH);
          break;
      default:
          lose = 1;
          break;
        break;
      }

    if (lose || optind < argc) {
        int i, space_left = KERNEL_PARAMAREA_LENGTH;
        binary_file = argv[optind];
        binary = 1;

        for (i = optind+1; i < argc && space_left > 0; i++) {
            space_left -= strlen(argv[i]) + 1;
            strncat(kernel_params, argv[i], space_left - 1);
            if (i+1 != argc)
                strncat(kernel_params, " ", 1);
        }
        printf("%sx\n", kernel_params);
    }

    if (h) {
      /* Print help info and exit.  */
      printf ("\
YAMS, Yet Another Machine Simulator\n\
Usage yams [-chsv] [--config file] [--script file] [--help] [--version]\n\
           [binary [opt1] .. [opt n]]\n\
  -c, --config file       Read configuration from 'file' instead\n\
  -h, --help              Print a summary of the options\n\
  -s, --script file       Read a script from 'file' before going to prompt\n\
  -v, --version           Print the version information\n\
");
      exit (0);
    }

    if (v) {
      printf("YAMS - Yet Another Machine Simulator %s \n", version);
      printf("\n");
      printf("\
Copyright (C) 2002 Juha Aatrokoski, Timo Lilja, Leena Salmela,\n\
  Teemu Takanen, Aleksi Virtanen\n\
\n\
There is NO warranty; not even for MERCHANTABILITY or FITNESS FOR A\n\
PARTICULAR PURPOSE.  You may redistribute copies of YAMS under the terms\n\
of the GNU General Public License.\n\
\n\
For more information about these matters, see the file named COPYING.\n");
      exit(0);
    }

    printf("YAMS - Yet Another Machine Simulator %s \n", version);
    printf("\n");

#ifdef WORDS_BIGENDIAN
    printf("Starting on a big endian host.\n");
#else
    printf("Starting on a little endian host.\n");
#endif /* WORDS_BIGENDIAN */

    printf("Simulating big endian machine.\n");

    {
        if (c) { /* read config from config_file */
            if (!cfg_read(config_file)) {
                printf(
"yams: unable to read configuration from file: %s\n", config_file);
                exit(1);
            }
        } else if (!cfg_read_cwd())
            ;
        else if (!cfg_read_home())
            ;
        else if (!cfg_read_etc())
            ;
        else {
            printf("\
yams: unable to read configuration file.\n");
            exit(1);
        }
        cfg_init();
    }

    {
        int ret = -1;
        simulator_init();

        if (binary) {
            command_boot(binary_file, kernel_params);
        }

        if (s) {
            int i;
            for (i = 0; i < scripts; i++) {
                ret = command_source(script_file[i]);
                if (ret == -2) /* file not found */
                    exit(1);
                else if (ret >= 0) /* stop yams */
                    break;
            }
        }
        if (ret == -1)
            exit_code = hwconsole_run();
	else
	    exit_code = ret;
    }

    exit(exit_code);
}


#ifdef C_ALLOCA /* xmalloc() is only used by alloca.c.  */

char *
xmalloc (size)
     unsigned int size;
{
  char *ptr = malloc (size);
  if (! ptr)
    {
      fprintf (stderr, _("%s: Virtual memory exhausted\n"), progname);
      exit (1);
    }
  return ptr;
}
#endif /* C_ALLOCA */

/* yams.c ends here */
